from typing import List

from clinica.pipelines.engine import Pipeline


class SpatialSVM(Pipeline):
    """SpatialSVM - Prepare input data for SVM with spatial and anatomical regularization.

    Returns:
        A clinica pipeline object containing the SpatialSVM pipeline.
    """

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        self.parameters.setdefault("group_label", None)
        check_group_label(self.parameters["group_label"])
        if "orig_input_data_ml" not in self.parameters.keys():
            raise KeyError(
                "Missing compulsory orig_input_data key in pipeline parameter."
            )
        # Optional parameters for inputs from pet-volume pipeline
        self.parameters.setdefault("acq_label", None)
        self.parameters.setdefault("suvr_reference_region", None)
        self.parameters.setdefault("use_pvc_data", False)
        # Advanced parameters
        self.parameters.setdefault("fwhm", 4)

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["dartel_input", "input_image"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list fo str :
            A list of (string) output fields name.
        """
        return ["regularized_image"]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.input_files import (
            pet_volume_normalized_suvr_pet,
            t1_volume_final_group_template,
        )
        from clinica.utils.inputs import clinica_file_reader, clinica_group_reader
        from clinica.utils.ux import print_groups_in_caps_directory

        if not (
            self.caps_directory / "groups" / f"group-{self.parameters['group_label']}"
        ).exists():
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                f"Group {self.parameters['group_label']} does not exist. "
                "Did you run pet-volume, t1-volume or t1-volume-create-dartel pipeline?"
            )

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )
        all_errors = []

        if self.parameters["orig_input_data_ml"] == "t1-volume":
            caps_files_information = {
                "pattern": os.path.join(
                    "t1",
                    "spm",
                    "dartel",
                    "group-" + self.parameters["group_label"],
                    "*_T1w_segm-graymatter_space-Ixi549Space_modulated-on_probability.nii.gz",
                ),
                "description": "graymatter tissue segmented in T1w MRI in Ixi549 space",
                "needed_pipeline": "t1-volume-tissue-segmentation",
            }
        elif self.parameters["orig_input_data_ml"] == "pet-volume":
            if not (
                self.parameters["acq_label"]
                and self.parameters["suvr_reference_region"]
            ):
                raise ValueError(
                    f"Missing value(s) in parameters from pet-volume pipeline. Given values:\n"
                    f"- acq_label: {self.parameters['acq_label']}\n"
                    f"- suvr_reference_region: {self.parameters['suvr_reference_region']}\n"
                    f"- use_pvc_data: {self.parameters['use_pvc_data']}\n"
                )
            caps_files_information = pet_volume_normalized_suvr_pet(
                acq_label=self.parameters["acq_label"],
                suvr_reference_region=self.parameters["suvr_reference_region"],
                use_brainmasked_image=False,
                use_pvc_data=self.parameters["use_pvc_data"],
                fwhm=0,
            )
        else:
            raise ValueError(
                f"Image type {self.parameters['orig_input_data_ml']} unknown."
            )

        try:
            input_image, _ = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                caps_files_information,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            dartel_input = clinica_group_reader(
                self.caps_directory,
                t1_volume_final_group_template(self.parameters["group_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        # Raise all errors if some happened
        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your CAPS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaCAPSError(error_message)

        read_parameters_node.inputs.dartel_input = dartel_input
        read_parameters_node.inputs.input_image = input_image

        self.connect(
            [
                (
                    read_parameters_node,
                    self.input_node,
                    [("dartel_input", "dartel_input")],
                ),
                (
                    read_parameters_node,
                    self.input_node,
                    [("input_image", "input_image")],
                ),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

        fisher_tensor_generation = npe.Node(
            name="obtain_g_fisher_tensor",
            interface=nutil.Function(
                input_names=["dartel_input", "FWHM"],
                output_names=["fisher_tensor", "fisher_tensor_path"],
                function=utils.obtain_g_fisher_tensor,
            ),
        )
        fisher_tensor_generation.inputs.FWHM = self.parameters["fwhm"]

        time_step_generation = npe.Node(
            name="estimation_time_step",
            interface=nutil.Function(
                input_names=["dartel_input", "FWHM", "g"],
                output_names=["t_step", "json_file"],
                function=utils.obtain_time_step_estimation,
            ),
        )
        time_step_generation.inputs.FWHM = self.parameters["fwhm"]

        heat_solver_equation = npe.MapNode(
            name="heat_solver_equation",
            interface=nutil.Function(
                input_names=["input_image", "g", "FWHM", "t_step", "dartel_input"],
                output_names=["regularized_image"],
                function=utils.heat_solver_equation,
            ),
            iterfield=["input_image"],
        )
        heat_solver_equation.inputs.FWHM = self.parameters["fwhm"]

        datasink = npe.Node(nio.DataSink(), name="sinker")
        datasink.inputs.base_directory = str(self.caps_directory)
        datasink.inputs.parameterization = True
        if self.parameters["orig_input_data_ml"] == "t1-volume":
            datasink.inputs.regexp_substitutions = [
                (
                    r"(.*)/regularized_image/.*/(.*(sub-(.*)_ses-(.*))_T1w(.*)_probability(.*))$",
                    r"\1/subjects/sub-\4/ses-\5/machine_learning/input_spatial_svm/group-"
                    + self.parameters["group_label"]
                    + r"/\3_T1w\6_spatialregularization\7",
                ),
                (
                    r"(.*)json_file/(output_data.json)$",
                    r"\1/groups/group-"
                    + self.parameters["group_label"]
                    + r"/machine_learning/input_spatial_svm/group-"
                    + self.parameters["group_label"]
                    + r"_space-Ixi549Space_parameters.json",
                ),
                (
                    r"(.*)fisher_tensor_path/(output_fisher_tensor.npy)$",
                    r"\1/groups/group-"
                    + self.parameters["group_label"]
                    + r"/machine_learning/input_spatial_svm/group-"
                    + self.parameters["group_label"]
                    + r"_space-Ixi549Space_gram.npy",
                ),
            ]

        elif self.parameters["orig_input_data_ml"] == "pet-volume":
            datasink.inputs.regexp_substitutions = [
                (
                    r"(.*)/regularized_image/.*/(.*(sub-(.*)_ses-(.*))_(task.*)_pet(.*))$",
                    r"\1/subjects/sub-\4/ses-\5/machine_learning/input_spatial_svm/group-"
                    + self.parameters["group_label"]
                    + r"/\3_\6_spatialregularization\7",
                ),
                (
                    r"(.*)json_file/(output_data.json)$",
                    r"\1/groups/group-"
                    + self.parameters["group_label"]
                    + r"/machine_learning/input_spatial_svm/group-"
                    + self.parameters["group_label"]
                    + r"_space-Ixi549Space_parameters.json",
                ),
                (
                    r"(.*)fisher_tensor_path/(output_fisher_tensor.npy)$",
                    r"\1/groups/group-"
                    + self.parameters["group_label"]
                    + r"/machine_learning/input_spatial_svm/group-"
                    + self.parameters["group_label"]
                    + r"_space-Ixi549Space_gram.npy",
                ),
            ]
        self.connect(
            [
                (
                    self.input_node,
                    fisher_tensor_generation,
                    [("dartel_input", "dartel_input")],
                ),
                (
                    fisher_tensor_generation,
                    time_step_generation,
                    [("fisher_tensor", "g")],
                ),
                (
                    self.input_node,
                    time_step_generation,
                    [("dartel_input", "dartel_input")],
                ),
                (
                    self.input_node,
                    heat_solver_equation,
                    [("input_image", "input_image")],
                ),
                (
                    fisher_tensor_generation,
                    heat_solver_equation,
                    [("fisher_tensor", "g")],
                ),
                (time_step_generation, heat_solver_equation, [("t_step", "t_step")]),
                (
                    self.input_node,
                    heat_solver_equation,
                    [("dartel_input", "dartel_input")],
                ),
                (
                    fisher_tensor_generation,
                    datasink,
                    [("fisher_tensor_path", "fisher_tensor_path")],
                ),
                (time_step_generation, datasink, [("json_file", "json_file")]),
                (
                    heat_solver_equation,
                    datasink,
                    [("regularized_image", "regularized_image")],
                ),
            ]
        )



"""Common CLI options used by Clinica tools."""

from multiprocessing import cpu_count

import click
from click_option_group import optgroup

option = optgroup.option

global_option_group = optgroup.group(
    "Options common to all clinica tools", help="Options common to all clinica tools"
)


def validate_n_procs(ctx, param, value):
    if isinstance(value, int) and 1 <= value <= cpu_count():
        return value
    raise click.BadParameter(
        "The number of processes should be between 1 and the number "
        f"of CPU available on your system ({cpu_count()}). "
        f"A value of {value} was provided instead. "
    )


n_procs = option(
    "-np",
    "--n_procs",
    type=int,
    default=lambda: max(cpu_count() - 1, 1),
    show_default="Number of available CPU minus one",
    callback=validate_n_procs,
    help="Number of cores used to run in parallel.",
)



import typing as ty

from nipype.interfaces.spm import Smooth
from pydra.engine import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task


def build_smoothing_workflow(
    name: str,
    fwhm: ty.Iterable[ty.Tuple[float, float, float]],
) -> Workflow:
    """Build and parametrize a smoothing workflow.

    Parameters
    ----------
    name : str
        The name of the Workflow.

    fwhm : List[List[float]]
        The smoothing kernel(s) to use for smoothing.
        If multiple kernels are provided, the workflow
        will output files for each specified kernel.
        Each file name has the kernel size added.

    Returns
    -------
    Workflow
        The resulting smoothing workflow.
    """
    wf = Workflow(
        name,
        input_spec=["input_file"],
    )
    outputs = []
    for x, y, z in fwhm:
        fwhm_name = f"{'-'.join([str(int(f)) for f in (x, y, z)])}mm"
        task = Nipype1Task(
            name=f"smoothing_node_{fwhm_name}",
            interface=Smooth(),
        )
        task.inputs.fwhm = fwhm_value
        task.inputs.out_prefix = f"fwhm-{fwhm_name}_"
        task.inputs.in_files = wf.lzin.input_file
        wf.add(task)
        outputs.append(
            (
                "smoothed_files",
                getattr(wf, f"smoothing_node_{fwhm_name}").lzout.smoothed_files,
            )
        )
    wf.set_output(outputs)
    return wf



"""This module contains Nipype tasks used by the connectome pipeline.

Nipype tasks must be 'self-contained' such that only primitive type hints can be
used. The tasks are simple wrappers around a properly implemented Python function.
"""


def convert_flirt_to_mrtrix_transformation_task(
    source_image: str,
    reference_image: str,
    flirt_matrix: str,
    name_output_matrix=None,
) -> str:
    """Adapter for Nipype"""

    from pathlib import Path

    from clinica.pipelines.dwi.connectome.utils import (
        convert_flirt_to_mrtrix_transformation,
    )

    if name_output_matrix is not None:
        name_output_matrix = Path(name_output_matrix)

    return str(
        convert_flirt_to_mrtrix_transformation(
            Path(source_image),
            Path(reference_image),
            Path(flirt_matrix),
            name_output_matrix,
        )
    )



from clinica.pipelines.engine import Pipeline
from clinica.utils.pet import ReconstructionMethod, Tracer
from clinica.utils.stream import log_and_raise


class PETPipeline(Pipeline):
    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.exceptions import ClinicaPipelineConfigurationError

        if "acq_label" not in self.parameters:
            log_and_raise(
                "Missing compulsory 'acq_label' key in pipeline parameter.",
                ClinicaPipelineConfigurationError,
            )
        self.parameters["acq_label"] = Tracer(self.parameters["acq_label"])
        if "reconstruction_method" in self.parameters:
            if self.parameters["reconstruction_method"]:
                self.parameters["reconstruction_method"] = ReconstructionMethod(
                    self.parameters["reconstruction_method"]
                )
        else:
            self.parameters["reconstruction_method"] = None

    def _get_pet_scans_query(self) -> dict:
        """Return the query to retrieve PET scans."""
        from clinica.utils.input_files import bids_pet_nii

        return bids_pet_nii(
            self.parameters["acq_label"], self.parameters["reconstruction_method"]
        )



from os import PathLike
from typing import List, Optional

__all__ = [
    "MissingModsTracker",
    "replace_sequence_chars",
    "write_longitudinal_analysis",
    "write_statistics",
]


def replace_sequence_chars(sequence_name: str) -> str:
    """Replace special characters in the sequence by underscores (as done for corresponding folder names in ADNI).

    Args:
        sequence_name: sequence to process

    Returns: the new string
    """
    import re

    return re.sub("[ /;*()<>:]", "_", sequence_name)


class MissingModsTracker:
    """Class used for tracking the number of missing modalities in a database.

    Attributes
    ----------
    missing : dict
        Dictionary of missing modalities.

    ses : list of str
        Sessions for which the tracker has information.
    """

    _default_modalities_tracked = ["dwi", "func", "fieldmap", "flair", "t1w"]

    def __init__(self, sessions: List[str], mod_list: Optional[List[str]] = None):
        modalities = mod_list or self._default_modalities_tracked
        if "session" not in modalities:
            modalities.insert(0, "session")
        self.missing = {s: {mod: 0 for mod in modalities} for s in sessions}

    @property
    def ses(self):
        return list(self.missing.keys())

    def add_missing_mod(self, session: str, modality: str) -> None:
        """Increase the number of missing files for the input modality.

        Parameters
        ----------
        session : str
            Name of the session for which to add a modality.

        modality : str
            The missing modality to add.
        """
        if session not in self.missing:
            raise ValueError(
                f"Session {session} was not provided to the MissingModsTracker constructor."
            )
        if modality not in self.missing[session]:
            raise ValueError(
                f"Modality {modality} is not tracked by this instance of MissingModsTracker."
            )
        self.missing[session][modality] += 1

    def increase_missing_ses(self, session: str) -> None:
        """Increase the session number.

        Parameters
        ----------
        session : str
            Name of the session.
        """
        if session not in self.missing:
            raise ValueError(
                f"Session {session} was not provided to the MissingModsTracker constructor."
            )
        self.missing[session]["session"] += 1

    def get_missing_list(self) -> dict:
        """Return the dictionary of missing modalities.

        Returns
        -------
        dict :
             The dictionary containing the list of missing files.
        """
        return self.missing


def sort_session_list(session_list: List[str]) -> List[str]:
    """Sorts the list of session IDs provided based on their session number.

    Parameters
    ------------
    session_list : list[str]
        List of session IDs to sort.

    Returns
    --------
    list[str] :
        Sorted list of session IDs.

    Examples
    --------
    >>>sort_session_list(["ses-M000", "ses-M006", "ses-M012", "ses-M024", "ses-M048", "ses-M003"])
    ["ses-M000", "ses-M003", "ses-M006", "ses-M012", "ses-M024", "ses-M048"]
    >>>sort_session_list(["ses-M0", "ses-M6", "ses-M12", "ses-M24", "ses-M48", "ses-M3"])
    ["ses-M0", "ses-M3", "ses-M6", "ses-M12", "ses-M24", "ses-M48"]
    """
    prefix_length = len("ses-M")
    session_idx = [
        ((session[prefix_length:]), int(session[prefix_length:]))
        for session in session_list
    ]
    session_idx.sort(key=lambda x: x[1])
    return [f"ses-M{session[0]}" for session in session_idx]


def write_statistics(
    summary_file: PathLike,
    num_subjs: int,
    ses_avail: List[str],
    mmt: MissingModsTracker,
) -> None:
    """Write statistics file.

    Write to a given input file statistics about missing files and modalities in a dataset.
    This method takes in input a MissingModsTracker object (mmt)
    that contains the number of missing modalities for each session and the
    number of missing sessions for each subject.

    Parameters
    ----------
    summary_file : PathLike
        Path of the output file where statistics should be written.

    num_subjs : int
        Number of subjects.

    ses_avail : list of str
        List of sessions available.

    mmt : MissingModsTracker
        Instance of MissingModsTracker.
    """
    with open(summary_file, "w") as fp:
        fp.write(compute_statistics(num_subjs, ses_avail, mmt))


def compute_statistics(
    num_subjs: int, ses_avail: List[str], mmt: MissingModsTracker
) -> str:
    """Compute statistics and return them as a string for printing/writing.

    Statistics are about missing files and modalities in a dataset.
    This method takes in input a MissingModsTracker object (mmt)
    that contains the number of missing modalities for each session and the
    number of missing sessions for each subject.

    Parameters
    ----------
    num_subjs : int
        Number of subjects.

    ses_avail : list of str
        List of sessions available.

    mmt : MissingModsTracker
        Instance of MissingModsTracker.

    Returns
    -------
    summary : str
        The statistics formatted as a summary string.
    """
    missing_list = mmt.get_missing_list()
    ses_avail = sort_session_list(ses_avail)
    ses_founds = {ses: num_subjs - missing_list[ses]["session"] for ses in ses_avail}
    summary = "\n".join(
        [
            "*" * 46,
            f"Number of subjects converted: {num_subjs}",
            f"Sessions available: {ses_avail}\n",
            "\n".join(
                [
                    f"Number of sessions {ses} found: {ses_found} ({ses_found * 100 / num_subjs}%)\n"
                    for ses, ses_found in ses_founds.items()
                ]
            ),
            "*" * 46 + "\n\n" + "Number of missing modalities for each session:\n",
        ]
    )
    for ses in ses_avail:
        summary += "\n" + ses + "\n"
        for mod in missing_list[ses]:
            if mod != "session":
                num_miss_mod = missing_list[ses][mod]
                percentage_missing = round(
                    (
                        num_miss_mod
                        * 100
                        / float(num_subjs - missing_list[ses]["session"])
                    ),
                    2,
                )
                summary += f"{mod}: {num_miss_mod} ({percentage_missing}%) \n"

    return summary


def write_longitudinal_analysis(
    summary_file: PathLike,
    bids_dir: PathLike,
    out_dir: PathLike,
    ses_avail: List[str],
    out_file_name: str,
) -> None:
    """Write to a given input file statistics about the present modalities and diagnoses in a dataset for each session.

    Parameters
    ----------
    summary_file : PathLike
        Path to the file where statistics should be written.

    bids_dir : PathLike
        Path to the BIDS directory.

    out_dir : PathLike
        Path to the output directory of the check-missing-modality pipeline.

    ses_avail : list of str
        List of sessions available.

    out_file_name : str
        String that replaces the default prefix ('missing_mods_') in
        the name of all the output files.
    """
    with open(summary_file, "w") as fp:
        fp.write(
            compute_longitudinal_analysis(bids_dir, out_dir, ses_avail, out_file_name)
        )


def compute_longitudinal_analysis(
    bids_dir: PathLike,
    out_dir: PathLike,
    ses_avail: List[str],
    out_file_name: str,
) -> str:
    """Compute statistics about the present modalities and diagnoses in a dataset for each session.

    Parameters
    ----------
    bids_dir : PathLike
        Path to the BIDS directory.

    out_dir : PathLike
        Path to the output directory of the check-missing-modality pipeline.

    ses_avail : list of str
        List of sessions available.

    out_file_name : str
        String that replaces the default prefix ('missing_mods_') in
        the name of all the output files.

    Returns
    -------
    summary : str
        The summary analysis as a string.
    """
    from collections import Counter
    from pathlib import Path

    import pandas as pd

    out_dir = Path(out_dir)
    bids_dir = Path(bids_dir)
    ses_avail = sort_session_list(ses_avail)
    summary = "\n\n".join(
        ["*" * 46, f"Number of present diagnoses and modalities for each session:\n"]
    )
    for ses in ses_avail:
        ses_df = pd.read_csv(
            out_dir / (out_file_name + ses + ".tsv"), sep="\t"
        ).set_index("participant_id")
        mods_avail = ses_df.columns.values
        mod_dict = dict()
        for mod in mods_avail:
            diagnosis_counter = Counter()
            subjects_avail = ses_df[ses_df[mod] == 1].index.values
            for subject in subjects_avail:
                subj_tsv_path = bids_dir / subject / f"{subject}_sessions.tsv"
                subj_df = pd.read_csv(subj_tsv_path, sep="\t").set_index("session_id")
                diagnosis = "missing"
                if (
                    ses in subj_df.index.values
                    and "diagnosis" in subj_df.columns.values
                ):
                    diagnosis = subj_df.loc[ses, "diagnosis"]
                    diagnosis = diagnosis if isinstance(diagnosis, str) else "n/a"
                diagnosis_counter.update([diagnosis])
            mod_dict[mod] = dict(diagnosis_counter)
        summary += f"{ses}\n{compute_table(mod_dict)}\n\n"
    return summary


def compute_table(mod_dict: dict) -> str:
    """Builds a table, encoded as a string, describing the
    modalities in the given dictionary.

    Parameters
    ----------
    mod_dict : dict
        Dictionary of modalities for which to compute the table.

    Returns
    -------
    table : str
        The summary table as a string.
    """
    diagnoses = sorted(set().union(*mod_dict.values()))
    table = "\t" + "\t| ".join(diagnoses)
    table += "\n" + "-" * 8 * (len(diagnoses) + 2) + "\n"
    for mod_key, mod_value in mod_dict.items():
        table += f"{mod_key}"
        if len(mod_key) < 8:
            table += "\t"
        table += "\t| ".join([str(mod_value.get(d, "0")) for d in diagnoses]) + "\n"
    return table


def viscode_to_session(viscode: str, baseline_identifiers: Optional[set] = None) -> str:
    """Replace the session label from 'baseline_identifiers' with 'ses-M000'.

    If `viscode` is not a baseline identifier, parse the `viscode` and return
    a session identifier of the form `ses-MXXX`.

    The `viscode` is expected to be formatted in the following way:
    "m/M{session_number}", where session_number can be casted to an integer.
    Otherwise, a `ValueError` will be raised.

    Note: This function doesn't perform any check on the number of digits
    of the session identifier. It will return at least three digits encoded
    session numbers, but doesn't raise if more digits are needed (see
    examples section below).

    Parameters
    ----------
    viscode: str
        The name of the session.

    baseline_identifiers : set of str, optional
        Possible identifiers for baseline session.
        If the `viscode` is among these identifiers, `ses-M000` is returned.
        Default={"bl", "m0"}.

    Returns
    -------
    str:
        "ses-M000" if the session is the baseline session.
        Otherwise returns the original session name capitalized.

    Raises
    ------
    ValueError :
        If the `viscode` isn't formatted as expected.

    Examples
    --------
    >>> viscode_to_session("m1")
    'ses-M001'
    >>> viscode_to_session("M123")
    'ses-M123'
    >>> viscode_to_session("m1234")
    'ses-M1234'
    """
    import re

    baseline_identifiers = baseline_identifiers or {"bl", "m0"}
    if viscode in baseline_identifiers:
        return "ses-M000"
    session_pattern = "^[m,M][0-9]*"
    if re.match(session_pattern, viscode):
        return "ses-" + f"M{(int(viscode[1:])):03d}"
    raise ValueError(
        f"The viscode {viscode} is not correctly formatted."
        "Expected a session identifier of the form 'MXXX', "
        f"or a baseline identifier among {baseline_identifiers}."
    )



"""This module contains utilities for longitudinal pipelines.

See CAPS specifications for details about long ID.
"""
from os import PathLike
from typing import List, Optional, Tuple, Union


def get_unique_subjects(
    subjects: List[str], sessions: List[str]
) -> Tuple[List[str], List[List[str]]]:
    """Get unique participant IDs with their sessions.

    This function generates a list of unique participant IDs from `in_subject_list` with their sessions.

    Parameters
    ----------
    subjects : list of str
        List of participant IDs.

    sessions : list of str
        List of session IDs.

    Returns
    -------
    unique_subjects : list of str
        List of participant IDs, where each participant appears only once.

    sessions_per_subject : list of lists of str
        List of session IDs associated to any single participant.

    Examples
    --------
    >>> from clinica.utils.participant import get_unique_subjects
    >>> get_unique_subjects(['sub-CLNC01', 'sub-CLNC01', 'sub-CLNC02'], ['ses-M000', 'ses-M018', 'ses-M000'])
    (['sub-CLNC01', 'sub-CLNC02'], [['ses-M000', 'ses-M018'], ['ses-M000']])
    """
    import numpy as np

    if len(subjects) != len(sessions):
        raise ValueError(
            "The number of subjects should match the number of sessions.\n"
            f"You provided the following subjects: {subjects}.\n"
            f"And the following sessions: {sessions}."
        )
    subjects = np.array(subjects)
    sessions = np.array(sessions)

    # The second returned element indicates for each participant_id the
    # element they correspond to in the 'unique' list. We will use this
    # to link each session_id in the repeated list of session_id to
    # their corresponding unique participant_id

    unique_subjects, inverse_positions = np.unique(subjects, return_inverse=True)
    sessions_per_subject = [
        sessions[inverse_positions == subject_index].tolist()
        for subject_index in range(len(unique_subjects))
    ]
    if len(unique_subjects) != len(sessions_per_subject):
        raise ValueError("Problem while getting unique subjects and sessions lists.")

    return unique_subjects.tolist(), sessions_per_subject


def unique_subjects_sessions_to_subjects_sessions(
    subjects: List[str],
    sessions_per_subject: List[List[str]],
) -> Tuple[List[str], List[str]]:
    """Do reverse operation of get_unique_subjects function.

    Parameters
    ----------
    subjects : list of str
        List of unique subject identifiers.

    sessions_per_subject : list of lists of str
        The sessions for each subject in subjects.

    Returns
    -------
    participants : list of str
        The list of subjects of same length as sessions.

    sessions : list of str
        The list of sessions of same length as participants.

    Examples
    --------
    >>> from clinica.utils.participant import unique_subjects_sessions_to_subjects_sessions
    >>> unique_subjects_sessions_to_subjects_sessions(['sub-01', 'sub-02'], [['ses-M000', 'ses-M018'], ['ses-M000']])
    (['sub-01', 'sub-01', 'sub-02'], ['ses-M000', 'ses-M018', 'ses-M000'])
    """
    if len(subjects) != len(sessions_per_subject):
        raise ValueError(
            "The number of unique subjects should match the number of session lists.\n"
            f"You provided the following subjects: {subjects}.\n"
            f"And the following session lists: {sessions_per_subject}."
        )
    participants, sessions = [], []
    for idx, participant_id in enumerate(subjects):
        for session_id in sessions_per_subject[idx]:
            participants.append(participant_id)
            sessions.append(session_id)

    return participants, sessions


def _read_participant_tsv(
    tsv_file: Union[str, PathLike],
) -> tuple[list[str], list[str]]:
    """Extract participant IDs and session IDs from TSV file.

    Parameters
    ----------
    tsv_file: str or Path
        Participant TSV file from which to extract the participant and session IDs.

    Returns
    -------
    participants: List[str]
        List of participant IDs.

    sessions: List[str]
        List of session IDs.

    Raises
    ------
    ClinicaException
        If `tsv_file` is not a file.
        If `participant_id` or `session_id` column is missing from TSV file.

    Examples
    --------
    >>> dframe = pd.DataFrame({
    ...     "participant_id": ["sub-01", "sub-01", "sub-02"],
    ...     "session_id": ["ses-M000", "ses-M006", "ses-M000"],
    ...})
    >>> dframe.to_csv("participants.tsv", sep="\t")
    >>> _read_participant_tsv("participant.tsv")
    (["sub-01", "sub-01", "sub-02"], ["ses-M000", "ses-M006", "ses-M000"])
    """
    import pandas as pd

    from clinica.utils.exceptions import ClinicaException

    try:
        df = pd.read_csv(tsv_file, sep="\t")
    except FileNotFoundError:
        raise ClinicaException(
            "The TSV file you gave is not a file.\nError explanations:\n"
            f"\t- Clinica expected the following path to be a file: {tsv_file}\n"
            "\t- If you gave relative path, did you run Clinica on the good folder?"
        )

    for column in ("participant_id", "session_id"):
        if column not in list(df.columns.values):
            raise ClinicaException(
                f"The TSV file does not contain {column} column (path: {tsv_file})"
            )

    return (
        [sub.strip(" ") for sub in list(df.participant_id)],
        [ses.strip(" ") for ses in list(df.session_id)],
    )


def get_subject_session_list(
    input_dir: PathLike,
    subject_session_file: Optional[PathLike] = None,
    is_bids_dir: bool = True,
    use_session_tsv: bool = False,
    tsv_dir: Optional[PathLike] = None,
) -> Tuple[List[str], List[str]]:
    """Parse a BIDS or CAPS directory to get the subjects and sessions.

    This function lists all the subjects and sessions based on the content of
    the BIDS or CAPS directory or (if specified) on the provided
    subject-sessions TSV file.

    Parameters
    ----------
    input_dir : PathLike
        A BIDS or CAPS directory path.

    subject_session_file : PathLike, optional
        A subjects-sessions file in TSV format.

    is_bids_dir : bool, optional
        Indicates if input_dir is a BIDS or CAPS directory.
        Default=True.

    use_session_tsv : bool, optional
        Specify if the list uses the sessions listed in the sessions.tsv files.
        Default=False.

    tsv_dir : PathLike, optional
        If TSV file does not exist, it will be created in output_dir.
        If not specified, output_dir will be in <tmp> folder

    Returns
    -------
    subjects : list of str
        Subjects list.

    sessions : list of str
        Sessions list.

    Notes
    -----
    This is a generic method based on folder names. If your <BIDS> dataset contains e.g.:
        - sub-CLNC01/ses-M000/anat/sub-CLNC01_ses-M000_T1w.nii
        - sub-CLNC02/ses-M000/dwi/sub-CLNC02_ses-M000_dwi.{bval|bvec|json|nii}
        - sub-CLNC02/ses-M000/anat/sub-CLNC02_ses-M000_T1w.nii
        get_subject_session_list(<BIDS>, None, True) will return
        ['ses-M000', 'ses-M000'], ['sub-CLNC01', 'sub-CLNC02'].

    However, if your pipeline needs both T1w and DWI files, you will need to check
    with e.g. clinica_file_reader_function.
    """
    import tempfile
    from pathlib import Path
    from time import localtime, strftime, time

    from clinica.iotools.utils.data_handling import create_subs_sess_list

    if not subject_session_file:
        output_dir = Path(tsv_dir) if tsv_dir else Path(tempfile.mkdtemp())
        timestamp = strftime("%Y%m%d_%H%M%S", localtime(time()))
        tsv_file = f"subjects_sessions_list_{timestamp}.tsv"
        subject_session_file = output_dir / tsv_file
        create_subs_sess_list(
            input_dir=input_dir,
            output_dir=output_dir,
            file_name=tsv_file,
            is_bids_dir=is_bids_dir,
            use_session_tsv=use_session_tsv,
        )

    return _read_participant_tsv(subject_session_file)



from os import PathLike
from typing import Iterable, Optional, Union

__all__ = ["center_nifti"]


def center_nifti(
    bids_directory: Union[str, PathLike],
    output_bids_directory: Union[str, PathLike],
    modalities: Optional[Iterable[str]] = None,
    center_all_files: bool = False,
    overwrite_existing_files: bool = False,
):
    """Center NIfTI files in a BIDS dataset.

    Parameters
    ----------
    bids_directory : PathLike
        The path to the BIDS directory.

    output_bids_directory : PathLike
        The path to

    modalities : Iterable of str, optional
        The modalities

    center_all_files : bool, optional
        Whether to center all file or not.
        Default=False.

    overwrite_existing_files : bool, optional
        If True and if the output BIDS directory already contain files,
        they might be overwritten. If False, the output BIDS has to be empty
        or non-existing otherwise a ClinicaExistingDatasetError will be raised.

    Notes
    -----
    This tool is mainly useful as a preprocessing step of SPM. In some cases, SPM is not able to segment T1 volumes
    because their respective center is not aligned with the origin of the world coordinate system. By default, only
    images detected as problematic are converted from INPUT_BIDS_DIRECTORY to OUTPUT_BIDS_DIRECTORY, whilst the others
    are copied verbatim.
    """
    import time
    from pathlib import Path

    from clinica.iotools.utils.data_handling import (
        center_all_nifti,
        write_list_of_files,
    )
    from clinica.utils.exceptions import ClinicaExistingDatasetError
    from clinica.utils.stream import cprint, log_and_raise

    bids_directory = Path(bids_directory)
    output_bids_directory = Path(output_bids_directory)
    if output_bids_directory.exists():
        files = [
            file.name
            for file in output_bids_directory.iterdir()
            if not file.name.startswith(".")
        ]
        if files and not overwrite_existing_files:
            raise ClinicaExistingDatasetError(output_bids_directory)
    cprint("Clinica is now centering the requested images.", lvl="info")

    centered_files = center_all_nifti(
        bids_directory,
        output_bids_directory,
        modalities,
        center_all_files,
    )
    # Write list of created files
    timestamp = time.strftime("%Y%m%d-%H%M%S", time.localtime(time.time()))
    log_file = output_bids_directory / f"centered_nifti_list_{timestamp}.txt"
    if not write_list_of_files(centered_files, log_file):
        log_and_raise(f"Could not create log file {log_file}.", IOError)

    cprint(
        f"{len(centered_files)} NIfTI files/images of BIDS folder:\n"
        f"\t{bids_directory}\n"
        f"for the modalities {modalities} have been centered in output folder:\n"
        f"\t{output_bids_directory}",
        lvl="info",
    )
    if log_file.is_file():
        cprint(
            f"The list of centered NIfTI files is available here: {log_file}.",
            lvl="info",
        )
    cprint(
        "Please note that the rest of the input BIDS folder has also been copied to the output folder.",
        lvl="info",
    )



__all__ = [
    "init_input_node_task",
    "move_subjects_dir_to_source_dir_task",
    "save_to_caps_task",
]


def init_input_node_task(
    caps_dir: str, participant_id: str, list_session_ids: list, output_dir: str
) -> tuple:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.template.utils import (
        init_input_node,
    )

    image_id, subjects_dir, flags = init_input_node(
        Path(caps_dir),
        participant_id,
        list_session_ids,
        Path(output_dir),
    )
    return image_id, str(subjects_dir), flags


def move_subjects_dir_to_source_dir_task(
    subjects_dir: str, source_dir: str, subject_id: str
):
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.utils import (
        move_subjects_dir_to_source_dir,
    )

    return move_subjects_dir_to_source_dir(
        Path(subjects_dir),
        Path(source_dir),
        subject_id,
    )


def save_to_caps_task(
    source_dir: str,
    image_id: str,
    list_session_ids: list,
    caps_dir: str,
    overwrite_caps: bool = False,
):
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.template.utils import (
        save_to_caps,
    )

    return save_to_caps(
        Path(source_dir),
        image_id,
        list_session_ids,
        Path(caps_dir),
        overwrite_caps=overwrite_caps,
    )



"""This module contains utilities used by the DWIDTI pipeline."""

from enum import Enum
from pathlib import Path
from typing import Dict, List, Tuple

__all__ = [
    "DTIBasedMeasure",
    "compute_statistics_on_atlases",
    "get_caps_filenames",
    "rename_into_caps",
    "print_begin_pipeline",
    "print_end_pipeline",
    "get_ants_transforms",
]


class DTIBasedMeasure(str, Enum):
    """Possible DTI measures."""

    FRACTIONAL_ANISOTROPY = "FA"
    MEAN_DIFFUSIVITY = "MD"
    AXIAL_DIFFUSIVITY = "AD"
    RADIAL_DIFFUSIVITY = "RD"


def compute_statistics_on_atlases(
    registered_map: Path, name_map: str, dwi_preprocessed_file: Path
) -> List[Path]:
    """Computes a list of statistics files for each atlas.

    Parameters
    ----------
    registered_map : Path
        Map already registered on atlases.

    name_map : str
        Name of the registered map in CAPS format.

    dwi_preprocessed_file : Path
        The preprocessed DWI file name which contains the entities to be
        used for building the statistics file names.

    Returns
    -------
    list of str :
        List of paths leading to the statistics TSV files.
    """
    from pathlib import Path

    from clinica.utils.atlas import atlas_factory
    from clinica.utils.bids import BIDSFileName
    from clinica.utils.statistics import statistics_on_atlas

    atlas_statistics_list = []
    for atlas_name in ("JHUDTI81", "JHUTracts0", "JHUTracts25"):
        atlas = atlas_factory(atlas_name)
        source = BIDSFileName.from_name(dwi_preprocessed_file)
        source.update_entity("space", atlas.name)
        source.update_entity("res", atlas.spatial_resolution)
        source.update_entity("map", name_map)
        source.suffix = "statistics"
        source.extension = ".tsv"
        out_atlas_statistics = (Path.cwd() / source.name).resolve()
        statistics_on_atlas(registered_map, atlas, out_atlas_statistics)
        atlas_statistics_list.append(out_atlas_statistics)

    return atlas_statistics_list


def get_caps_filenames(caps_dwi_filename: Path) -> Dict[str, str]:
    """Prepare some filenames with CAPS naming convention."""
    import re

    m = re.search(
        r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_space-[a-zA-Z0-9]+_desc-preproc",
        str(caps_dwi_filename),
    )
    if not m:
        raise ValueError(
            f"Input filename {caps_dwi_filename} is not in a CAPS compliant format."
        )

    caps_prefix = m.group(0)
    results = {
        "bids_source": f"{m.group(1)}_{m.group(2)}",
        "diffmodel": f"{caps_prefix}_model-DTI_diffmodel.nii.gz",
    }
    results.update(
        {
            measure.value: f"{caps_prefix}_{measure.value}.nii.gz"
            for measure in DTIBasedMeasure
        }
    )
    results["DECFA"] = f"{caps_prefix}_DECFA.nii.gz"

    return results


def rename_into_caps(
    in_caps_dwi: Path,
    in_norm_fa: Path,
    in_norm_md: Path,
    in_norm_ad: Path,
    in_norm_rd: Path,
    in_b_spline_transform: Path,
    in_affine_matrix: Path,
) -> Tuple[str, ...]:
    """Rename different outputs of the pipelines into CAPS format.

    Parameters
    ----------
    in_caps_dwi : Path
    in_norm_fa : Path
    in_norm_md : Path
    in_norm_ad : Path
    in_norm_rd : Path
    in_b_spline_transform : Path
    in_affine_matrix : Path

    Returns
    -------
    tuple of str :
        The different outputs with CAPS naming convention
    """
    from clinica.pipelines.dwi.utils import rename_files

    return rename_files(
        in_caps_dwi,
        {
            in_norm_fa: "_space-MNI152Lin_res-1x1x1_FA.nii.gz",
            in_norm_md: "_space-MNI152Lin_res-1x1x1_MD.nii.gz",
            in_norm_ad: "_space-MNI152Lin_res-1x1x1_AD.nii.gz",
            in_norm_rd: "_space-MNI152Lin_res-1x1x1_RD.nii.gz",
            in_b_spline_transform: "_space-MNI152Lin_res-1x1x1_deformation.nii.gz",
            in_affine_matrix: "_space-MNI152Lin_res-1x1x1_affine.mat",
        },
    )


def print_begin_pipeline(in_bids_or_caps_file: str):
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    print_begin_image(get_subject_id(in_bids_or_caps_file))


def print_end_pipeline(in_bids_or_caps_file: str, final_file_1: str, final_file_2: str):
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(in_bids_or_caps_file))


def get_ants_transforms(
    in_affine_transformation: str, in_bspline_transformation: str
) -> list:
    """Combine transformations for antsApplyTransforms interface."""
    return [in_bspline_transformation, in_affine_transformation]



from typing import List

from clinica.pipelines.engine import Pipeline


class StatisticsVolumeCorrection(Pipeline):
    """StatisticsVolumeCorrection - Statistical correction of StatisticsVolume pipeline.

    Returns:
        A clinica pipeline object containing the StatisticsVolumeCorrection pipeline.
    """

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["t_map"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return []

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.inputs import clinica_group_reader

        t_map = clinica_group_reader(
            self.caps_directory,
            {
                "pattern": self.parameters["t_map"] + "*",
                "description": "statistics t map",
                "needed_pipeline": "statistics-volume",
            },
        )

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )
        read_parameters_node.inputs.t_map = t_map

        self.connect([(read_parameters_node, self.input_node, [("t_map", "t_map")])])

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        from os.path import abspath, dirname, exists, join, pardir

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import numpy as np

        import clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils as utils
        from clinica.utils.inputs import RemoteFileStructure, fetch_file
        from clinica.utils.stream import cprint

        peak_correction_FWE = npe.Node(
            name="peak_correction_FWE",
            interface=nutil.Function(
                input_names=["t_map", "t_threshold"],
                output_names=["output"],
                function=utils.peak_correction,
            ),
        )
        peak_correction_FWE.inputs.t_threshold = self.parameters["FWEp"]

        peak_correction_FDR = peak_correction_FWE.clone(name="peak_correction_FDR")
        peak_correction_FDR.inputs.t_threshold = self.parameters["FDRp"]

        cluster_correction_FWE = npe.Node(
            name="cluster_correction_FWE",
            interface=nutil.Function(
                input_names=["t_map", "t_thresh", "c_thresh"],
                output_names=["output"],
                function=utils.cluster_correction,
            ),
        )
        cluster_correction_FWE.inputs.t_thresh = self.parameters["height_threshold"]
        cluster_correction_FWE.inputs.c_thresh = self.parameters["FWEc"]

        cluster_correction_FDR = cluster_correction_FWE.clone(
            name="cluster_correction_FDR"
        )
        cluster_correction_FDR.inputs.t_thresh = self.parameters["height_threshold"]
        cluster_correction_FDR.inputs.c_thresh = self.parameters["FDRc"]

        produce_fig_FWE_peak_correction = npe.Node(
            name="produce_figure_FWE_peak_correction",
            interface=nutil.Function(
                input_names=[
                    "nii_file",
                    "template",
                    "type_of_correction",
                    "t_thresh",
                    "c_thresh",
                    "n_cuts",
                ],
                output_names=["figs"],
                function=utils.produce_figures,
            ),
        )
        produce_fig_FWE_peak_correction.inputs.n_cuts = self.parameters["n_cuts"]

        root = dirname(abspath(join(abspath(__file__), pardir, pardir)))
        path_to_mask = join(root, "resources", "masks")
        url_aramis = "https://aramislab.paris.inria.fr/files/data/img_t1_linear/"
        FILE1 = RemoteFileStructure(
            filename="mni_icbm152_t1_tal_nlin_sym_09a.nii.gz",
            url=url_aramis,
            checksum="3b244ee7e287319d36a25263744c468ef0ab2fe5a94b15a2138844db73b49adf",
        )

        produce_fig_FWE_peak_correction.inputs.template = join(
            path_to_mask, FILE1.filename
        )
        if not (exists(produce_fig_FWE_peak_correction.inputs.template)):
            try:
                fetch_file(FILE1, path_to_mask)
            except IOError as err:
                cprint(
                    msg=f"Unable to download required template (mni_icbm152) for processing: {err}",
                    lvl="error",
                )

        produce_fig_FDR_peak_correction = produce_fig_FWE_peak_correction.clone(
            name="produce_figure_FDR_peak_correction"
        )
        produce_fig_FWE_cluster_correction = produce_fig_FWE_peak_correction.clone(
            name="produce_figure_FWE_cluster_correction"
        )
        produce_fig_FDR_cluster_correction = produce_fig_FWE_peak_correction.clone(
            name="produce_figure_FDR_cluster_correction"
        )

        # fmt: off
        produce_fig_FWE_peak_correction.inputs.type_of_correction = "FWE"
        produce_fig_FDR_peak_correction.inputs.type_of_correction = "FDR"
        produce_fig_FWE_cluster_correction.inputs.type_of_correction = "FWE"
        produce_fig_FDR_cluster_correction.inputs.type_of_correction = "FDR"

        produce_fig_FWE_peak_correction.inputs.t_thresh = self.parameters["FWEp"]
        produce_fig_FDR_peak_correction.inputs.t_thresh = self.parameters["FDRp"]
        produce_fig_FWE_cluster_correction.inputs.t_thresh = self.parameters["height_threshold"]
        produce_fig_FDR_cluster_correction.inputs.t_thresh = self.parameters["height_threshold"]

        produce_fig_FWE_peak_correction.inputs.c_thresh = np.nan
        produce_fig_FDR_peak_correction.inputs.c_thresh = np.nan
        produce_fig_FWE_cluster_correction.inputs.c_thresh = self.parameters["FWEc"]
        produce_fig_FDR_cluster_correction.inputs.c_thresh = self.parameters["FDRc"]
        # fmt: on

        save_fig_peak_correction_FWE = npe.Node(
            name="save_figure_peak_correction_FWE",
            interface=nutil.Function(
                input_names=["t_map", "figs", "correction_name"],
                output_names=[],
                function=utils.generate_output,
            ),
        )
        save_fig_peak_correction_FWE.inputs.correction_name = "FWEp"

        save_fig_peak_correction_FDR = save_fig_peak_correction_FWE.clone(
            name="save_fig_peak_correction_FDR"
        )
        save_fig_peak_correction_FDR.inputs.correction_name = "FDRp"

        save_fig_cluster_correction_FWE = save_fig_peak_correction_FWE.clone(
            name="save_fig_cluster_correction_FWE"
        )
        save_fig_cluster_correction_FWE.inputs.correction_name = "FWEc"

        save_fig_cluster_correction_FDR = save_fig_peak_correction_FWE.clone(
            name="save_fig_cluster_correction_FDR"
        )
        save_fig_cluster_correction_FDR.inputs.correction_name = "FDRc"

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, peak_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, peak_correction_FDR, [("t_map", "t_map")]),
                (self.input_node, cluster_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, cluster_correction_FDR, [("t_map", "t_map")]),
                (peak_correction_FWE, produce_fig_FWE_peak_correction, [("output", "nii_file")]),
                (peak_correction_FDR, produce_fig_FDR_peak_correction, [("output", "nii_file")]),
                (cluster_correction_FWE, produce_fig_FWE_cluster_correction, [("output", "nii_file")]),
                (cluster_correction_FDR, produce_fig_FDR_cluster_correction, [("output", "nii_file")]),
                (produce_fig_FWE_peak_correction, save_fig_peak_correction_FWE, [("figs", "figs")]),
                (produce_fig_FDR_peak_correction, save_fig_peak_correction_FDR, [("figs", "figs")]),
                (produce_fig_FWE_cluster_correction, save_fig_cluster_correction_FWE, [("figs", "figs")]),
                (produce_fig_FDR_cluster_correction, save_fig_cluster_correction_FDR, [("figs", "figs")]),
                (self.input_node, save_fig_peak_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, save_fig_peak_correction_FDR, [("t_map", "t_map")]),
                (self.input_node, save_fig_cluster_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, save_fig_cluster_correction_FDR, [("t_map", "t_map")]),
            ]
        )



from typing import Any

import pydra
from nipype.algorithms.misc import Gunzip
from nipype.interfaces.spm import DARTELNorm2MNI
from pydra import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pydra.engine import clinica_io


@clinica_io
def build_core_workflow(
    name: str = "t1-volume-dartel2mni", parameters: dict = {}
) -> Workflow:
    """Core workflow for T1VolumeDartel2MNI - Dartel template to MNI (Pydra engine).

    Parameters
    ----------
    name : str, optional
        Name of the workflow. Default="t1-volume-dartel2mni".

    parameters : dict, optional
        Dictionary of parameters. Default={}.
    """
    from clinica.pydra.shared_workflows.smoothing import build_smoothing_workflow
    from clinica.pydra.t1_volume.dartel2mni.tasks import prepare_flowfields_task
    from clinica.pydra.utils import sanitize_fwhm
    from clinica.utils.spm import use_spm_standalone_if_available

    use_spm_standalone_if_available()

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", Any),
            (
                "tissues",
                dict,
                {"tissue_number": parameters["tissues"]},
                {"mandatory": False},
            ),
            (
                "flow_fields",
                dict,
                {"group_label": parameters["group_label"]},
                {"mandatory": True},
            ),
            (
                "dartel_template",
                dict,
                {"group_label": parameters["group_label"]},
                {"mandatory": True},
            ),
        ],
        bases=(pydra.specs.BaseSpec,),
    )

    wf = Workflow(name, input_spec=input_spec)

    wf.split(("tissues", "flow_fields"))

    compressed_inputs = [
        "flow_fields",
        "dartel_template",
    ]

    # Unzipping
    for input_name in compressed_inputs:
        wf.add(
            Nipype1Task(
                name=f"unzip_{input_name}",
                interface=Gunzip(),
                in_file=getattr(wf.lzin, input_name),
            )
        )

    wf.add(
        Nipype1Task(
            name="unzip_tissues",
            interface=Gunzip(),
            in_file=wf.lzin.tissues,
        )
        .split("in_file")
        .combine("in_file")
    )

    # DARTEL2MNI Registration
    wf.add(
        prepare_flowfields_task(
            name="prepare_flowfields",
            interface=prepare_flowfields_task,
            flowfields=wf.unzip_flow_fields.lzout.out_file,
            number_of_tissues=len(parameters["tissues"]),
        )
    )

    dartel2mni_node = Nipype1Task(
        name="dartel2mni",
        interface=DARTELNorm2MNI(),
    )
    if parameters["voxel_size"] is not None:
        dartel2mni_node.inputs.voxel_size = tuple(parameters["voxel_size"])
    dartel2mni_node.inputs.modulate = parameters["modulate"]
    dartel2mni_node.inputs.fwhm = 0
    dartel2mni_node.inputs.apply_to_files = wf.unzip_tissues.lzout.out_file
    dartel2mni_node.inputs.template_file = wf.unzip_dartel_template.lzout.out_file
    dartel2mni_node.inputs.flowfield_files = (
        wf.prepare_flowfields.lzout.prepared_flowfields
    )
    wf.add(dartel2mni_node)

    output_connections = [("normalized_files", wf.dartel2mni.lzout.normalized_files)]

    # Smoothing
    if parameters["smooth"] is not None:
        smoothing_wf = build_smoothing_workflow(
            "smoothing_workflow",
            sanitize_fwhm(parameters["smooth"]),
        )
        smoothing_wf.inputs.input_file = wf.dartel2mni.lzout.normalized_files
        wf.add(smoothing_wf)
        output_connections += [
            ("smoothed_normalized_files", wf.smoothing_workflow.lzout.smoothed_files),
        ]

    wf.set_output(output_connections)

    return wf



from typing import Optional

import numpy as np

__all__ = ["Results"]


def _convert_arrays_to_lists(data: dict) -> dict:
    """If the input dictionary contains numpy arrays, this function will
    cast them to lists and return the same dictionary with the lists instead
    of the numpy arrays.

    Parameters
    ----------
    data : dict
        The dictionary to clean.

    Returns
    -------
    new_data : dict
        The dictionary with arrays casted to lists.
    """
    new_data = dict()
    for k, v in data.items():
        if isinstance(v, dict):
            new_data[k] = _convert_arrays_to_lists(v)
        elif isinstance(v, np.ndarray):
            new_data[k] = v.tolist()
        else:
            new_data[k] = v
    return new_data


class Results:
    """Common class for GLM results."""

    def to_dict(self) -> dict:
        """Returns the `Results` instance in dict format.

        Private attributes and all methods are not returned.

        This function does not perform any casting.

        Returns
        -------
        data : dict
            Resulting dictionary.
        """
        import inspect

        data = dict()
        for attribute in inspect.getmembers(self):
            name, value = attribute
            if not name.startswith("_"):
                if not inspect.ismethod(value):
                    if hasattr(value, "to_dict"):
                        data[name] = value.to_dict()
                    else:
                        data[name] = value
        return data

    def to_json(self, indent: Optional[int] = 4) -> str:
        """Returns the json of the `Results` instance.

        Parameters
        ----------
        indent : int, optional
            Indent to use. Default=4.

        Returns
        -------
        str :
            The JSON dumps of the results.
        """
        import json

        return json.dumps(_convert_arrays_to_lists(self.to_dict()), indent=indent)
